#coding: utf-8

"""Offers "Log in via oDesk" functionality."""

O_REQUEST_TOKEN = '_ort'
"""Session key under which OAuth request token will be stored
during authentication."""

O_ACCESS_TOKEN = '_oat'
"""Session key under which OAuth access token will be stored
once authentication is complete."""



#coding: utf-8

from django.core.urlresolvers import reverse
from django import http
from django.contrib.auth import authenticate, login
from django.contrib.sites.models import Site

from . import utils
from . import O_REQUEST_TOKEN, O_ACCESS_TOKEN


def oauth_login(request):
    """Initiates OAuth login process."""

    odesk_client = utils.get_client()

    # After oDesk redirects user back, we can fetch request token from session
    # and use it to obtain access token and finish authentication
    request.session[O_REQUEST_TOKEN] = odesk_client.auth.get_request_token()

    # (OAuth.get_request_token() magically sets ``request_token[_secret]``
    # attributes on OAuth instance under c.auth, so we don't have to do that)

    site = Site.objects.get_current()

    return http.HttpResponseRedirect(odesk_client.auth.get_authorize_url(
        'http://{domain}{url}?next={redirect}'.format(
            domain=site.domain,
            url=reverse('odesk_oauth_login_callback'),
            redirect=request.GET.get('next'),
        )
    ))


def oauth_login_callback(request):
    """Handles redirect from oDesk side during OAuth login process."""

    odesk_client = utils.get_client()

    if O_REQUEST_TOKEN in request.session:
        (
            odesk_client.auth.request_token,
            odesk_client.auth.request_token_secret,
        ) = request.session.get(O_REQUEST_TOKEN, [None, None])
        del request.session[O_REQUEST_TOKEN]

    else:
        return http.HttpResponse("Request token is missing", status=401)

    try:
        access_token = \
            odesk_client.auth.get_access_token(
                request.GET.get('oauth_verifier'))
    except Exception, e:
        return http.HttpResponse(e, status=401)

    user = authenticate(access_token=access_token)
    # ODeskOAuthBackend will retrieve user details
    # and create new user if it's first-time login

    if user:
        # If we want to make a request to oDesk API on user's behalf later
        # after authentication, access token tuple can be obtained
        # from user's session (and applied to corresponding
        # Client.auth.oauth_ attributes)
        request.session[O_ACCESS_TOKEN] = access_token

        if user.is_active:
            login(request, user)
        else:
            return http.HttpResponse("User is not active", status=401)
    else:
        return http.HttpResponse("User not found", status=401)

    redirect_url = request.GET.get('next', '/')

    # oDesk apparently sets "next" in GET query to string None
    if redirect_url == 'None':
        redirect_url = '/'

    return http.HttpResponseRedirect(redirect_url)


def oauth_check_login(request):
    """Verifies that there's a valid OAuth access token stored in user's
    session, and that user is active. If there's a problem,
    returns HTTP 401 (Unauthorized). Otherwise returns HTTP 200 (OK).
    Additional information might be in response body text.
    """
    access_token = request.session.get(O_ACCESS_TOKEN, None)

    if access_token is None or len(access_token) != 2:
        return http.HttpResponse("Access token is missing", status=401)

    result, details = utils.check_login(access_token)

    if result is True:
        return http.HttpResponse(details, status=200)
    else:
        return http.HttpResponse(details, status=401)



#coding: utf-8

from django.conf.urls import patterns, url

from . import views


urlpatterns = patterns('',
    url(r'^authenticate/', views.oauth_login,
        name='odesk_oauth_login'),
    url(r'^callback/', views.oauth_login_callback,
        name='odesk_oauth_login_callback'),
    url(r'^check-login/', views.oauth_check_login,
        name='odesk_oauth_check_login'),
)



#coding: utf-8

import logging
import urllib2
from odesk import Client

from odesk import exceptions as odesk_exceptions
from django.conf import settings
from django.core.exceptions import ImproperlyConfigured


logger = logging.getLogger(__name__)


def get_client(access_token=None):
    """Returns OAuth-enabled python-odesk Client instance."""
    try:
        key = settings.ODESK_OAUTH_KEY
        secret = settings.ODESK_OAUTH_SECRET
    except AttributeError:
        raise ImproperlyConfigured(
            "ODESK_OAUTH_KEY and/or ODESK_OAUTH_SECRET settings missing")

    odesk_client = Client(key, secret)

    if access_token:
        (
            odesk_client.oauth_access_token,
            odesk_client.oauth_access_token_secret,
        ) = access_token

    return odesk_client


def set_user_info(user, odesk_user):
    """Updates given ``user`` information (first and last name, email)
    based on ``odesk_user`` dictionary and saves the object.

    :param user: :class:`django.contrib.auth.models.User` instance
    :param odesk_user: a dictionary containing authenticated user information
                       received from oDesk HR API at /users/me
    """
    user.first_name = odesk_user['first_name']
    user.last_name = odesk_user['last_name']
    user.email = odesk_user['email']
    user.save()

    return user


def update_user_permissions(user, teams):
    """Updates given ``user`` permission-related attributes
    based on provided team list and project settings. Saves the object.

    :param user: :class:`django.contrib.auth.models.User` instance
    :param teams: names of oDesk teams that user belongs to
    """
    teams = set(teams)

    user.is_staff = any([
        teams.intersection(settings.ODESK_AUTH_ADMIN_TEAMS),
        user.username in settings.ODESK_AUTH_ADMINS,
    ])

    user.is_superuser = any([
        teams.intersection(settings.ODESK_AUTH_SUPERUSER_TEAMS),
        user.username in settings.ODESK_AUTH_SUPERUSERS,
    ])

    user.is_active = any([
        user.is_staff,
        user.is_superuser,
        user.username in settings.ODESK_AUTH_ALLOWED_USERS,
        teams.intersection(settings.ODESK_AUTH_ALLOWED_TEAMS),
    ])

    user.save()

    return user


def check_login(access_token):
    """Verifies that given OAuth ``access_token`` is valid, and user is active
    on oDesk. Makes a test request to oDesk API for that purpose.

    Returns a 2-tuple. If login is bad, it's ``False`` and a string with
    additional information. Otherwise it's ``(True, "OK")``.
    """
    odesk_client = get_client()
    (
        odesk_client.oauth_access_token,
        odesk_client.oauth_access_token_secret,
    ) = access_token

    try:
        auth_user_resp = odesk_client.hr.get_user('me')
    except odesk_exceptions.HTTP403ForbiddenError:
        return False, "Invalid access token"
    except (urllib2.HTTPError, urllib2.URLError):
        logger.exception("unexpected network error in check_login()")
        return False, "Network error"
    except ValueError as exc:
        logger.exception(exc)
        return False, "Value error"

    if auth_user_resp.get('status') != 'active':
        return False, "User is inactive"

    return True, "OK"



#coding: utf-8

import logging
from odesk import exceptions as odesk_exceptions

from django.contrib.auth.models import User
from django.conf import settings

from . import utils


logger = logging.getLogger(__name__)


CREATE_UNKNOWN_USER = getattr(settings, 'ODESK_AUTH_CREATE_UNKNOWN_USER', False)
"""Whether to create a new Django account if user logs in via oDesk
the first time."""

ODESK_USERNAME_SUFFIX = "@odesk.com"
"""String appended to oDesk user ID to be used
as :attr:`User.username <django.contrib.auth.models.User.username>`
for newly created users."""


class ODeskOAuthBackend(object):
    """Backend for Django auth system."""

    def authenticate(self, access_token=None):
        """Verifies that given OAuth ``access_token`` is valid by making
        a request to oDesk API to get current user's information.

        If request is successful, creates/updates corresponding
        :class:`~django.contrib.auth.models.User` object.

        **When user logs in for the first time,**
        :attr:`~django.contrib.auth.models.User.username` is set in the form
        ``'<odesk_username>@odesk.com'``, same as ``email``.
        ``first_name`` and ``last_name`` are also set based on user's info
        returned by oDesk API.

        .. seealso:: :func:`django_odesk_auth.utils.set_user_info`

        **On each subsequent logins,** user's permissions
        (``is_staff``, ``is_superuser``, ``is_active`` flags)
        are updated based on oDesk teams user belongs to and teams
        in FIVS :mod:`settings`.

        .. seealso:: :func:`django_odesk_auth.utils.update_user_permissions`
        """

        odesk_client = utils.get_client()
        (
            odesk_client.oauth_access_token,
            odesk_client.oauth_access_token_secret,
        ) = access_token

        try:
            auth_user = odesk_client.hr.get_user('me')
        except odesk_exceptions.HTTP403ForbiddenError:
            logger.exception(
                "Invalid access token given to authenticate(): %s",
                access_token)
            return None
        else:
            if auth_user.get('status') != 'active':
                logger.warn(
                    "Inactive user authentication attempt: %s",
                    auth_user.get('id'))
                return None

        username = "{odesk_id}{username_suffix}".format(
            odesk_id=auth_user['id'],
            username_suffix=ODESK_USERNAME_SUFFIX)

        try:
            user = User.objects.get(username=username)
        except User.DoesNotExist:
            if CREATE_UNKNOWN_USER:
                logger.info("Creating new user: %s", username)
                user = User.objects.create(username=username)
                user = utils.set_user_info(user, auth_user)
            else:
                logger.info(
                    "Unknown user %s tried to log in, declining",
                    username)
                return None

        teams = set(team['id'] for team in odesk_client.hr.get_teams())
        user = utils.update_user_permissions(user, teams)

        return user

    def get_user(self, user_id):
        return User.objects.get(id=user_id)



#coding: utf-8

import mock
import urllib2

import odesk

from django.test import TestCase

from . import utils


class LoginCheckTestCase(TestCase):

    @mock.patch('django_odesk_auth.utils.get_client')
    def test_http_error(self, client_getter):
        client = client_getter.return_value
        client.hr.get_user.side_effect = urllib2.HTTPError(
            '/fake-url',
            503,
            'Service Unavailable',
            {},  # headers
            None,  # fp
        )

        result = utils.check_login(('dummy', 'dummy'))

        self.assertTrue(client_getter.called)
        self.assertEqual(result, (False, "Network error"))

    @mock.patch('django_odesk_auth.utils.get_client')
    def test_url_error(self, client_getter):
        client = client_getter.return_value
        client.hr.get_user.side_effect = urllib2.URLError(
            'connection reset by peer')

        result = utils.check_login(('dummy', 'dummy'))

        self.assertTrue(client_getter.called)
        self.assertEqual(result, (False, "Network error"))

    @mock.patch('django_odesk_auth.utils.get_client')
    def test_invalid_token(self, client_getter):
        client = client_getter.return_value
        client.hr.get_user.side_effect = odesk.exceptions.HTTP403ForbiddenError(
            '/fake-url',
            403,
            'Forbidden',
            {},  # headers
            None,  # fp
        )

        result = utils.check_login(('dummy', 'dummy'))

        self.assertTrue(client_getter.called)
        self.assertEqual(result, (False, "Invalid access token"))

    @mock.patch('django_odesk_auth.utils.get_client')
    def test_inactive_user(self, client_getter):
        client = client_getter.return_value
        client.hr.get_user.return_value = {
            'status': 'bad status',
            'first_name': 'name',
            'last_name': 'surname',
            'uid': 'userid',
            'mail': 'user@example.net',
        }

        result = utils.check_login(('dummy', 'dummy'))

        self.assertTrue(client_getter.called)
        self.assertEqual(result, (False, "User is inactive"))

    @mock.patch('django_odesk_auth.utils.get_client')
    def test_ok(self, client_getter):
        client = client_getter.return_value
        client.hr.get_user.return_value = {
            'status': 'active',
            'first_name': 'name',
            'last_name': 'surname',
            'uid': 'userid',
            'mail': 'user@example.net',
        }

        result = utils.check_login(('dummy', 'dummy'))

        self.assertTrue(client_getter.called)
        self.assertEqual(result, (True, "OK"))
